// file      : xsde/cxx/parser/parser-inline.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-inline.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (ename (l));
          SemanticGraph::Type& t (l.argumented ().type ());

          String item (unclash (name, "item"));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // item_parser
          //
          os << inl
             << "void " << name << "::" << endl
             << unclash (name, "item_parser") << " (" <<
            fq_name (t) << "& " << item << ")"
             << "{"
             << "this->_xsde_" << item << "_ = &" << item << ";"
             << "}";

          // parsers
          //
          os << inl
             << "void " << name << "::" << endl
             << "parsers (" << fq_name (t) << "& " << item << ")"
             << "{"
             << "this->_xsde_" << item << "_ = &" << item << ";"
             << "}";

          // c-tor
          //
          os << inl
             << name << "::" << endl
             << name << " ()" << endl
             << ": _xsde_" << item << "_ (0)"
             << "{"
             << "}";
        }
      };

      //
      //
      struct ParserAccessor: Traversal::Member,
                             protected virtual Context
      {
        ParserAccessor (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));

          os << inl
             << "void " << ename (m.scope ()) << "::" << endl
             << eparser (m) << " (" << fq_name (m.type ()) << "& " <<
            name << ")"
             << "{"
             << "this->" << emember (m) << " = &" << name << ";"
             << "}";
        }
      };


      //
      //
      struct ParserMemberSet: Traversal::Member,
                              protected virtual Context
      {
        ParserMemberSet (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));

          os << "this->" << emember (m) << " = &" << name << ";";
        }
      };


      //
      //
      struct ParserMemberInit: Traversal::Member,
                               protected virtual Context
      {
        ParserMemberInit (Context& c)
            : Context (c), first_ (true)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          if (first_)
            first_ = false;
          else
            os << "," << endl << "  ";

          os << emember (m) << " (0)";
        }

        Boolean
        comma () const
        {
          return !first_;
        }

      private:
        Boolean first_;
      };

      struct ParserBaseSet: Traversal::Complex,
                            Traversal::List,
                            protected virtual Context
      {
        ParserBaseSet (Context& c)
            : Context (c), member_ (c)
        {
          inherits_ >> *this;
          names_ >> member_;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          inherits (c, inherits_);

          if (!restriction_p (c))
            names (c, names_);
        }

        virtual Void
        traverse (SemanticGraph::List& l)
        {
          String const& name (ename (l));
          String item (unclash (name, "item"));

          os << "this->_xsde_" << item << "_ = &" << name << "_item;";
        }

      private:
        Traversal::Inherits inherits_;

        ParserMemberSet member_;
        Traversal::Names names_;
      };


      struct Particle: Traversal::All,
                       protected virtual Context
      {
        Particle (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (SemanticGraph::All& a)
        {
          if (!a.context().count ("comp-number"))
            return;

          UnsignedLong state_count (
            a.context().get<UnsignedLong> ("state-count"));

          os << "," << endl
             << "  v_all_count_ (" << state_count << "UL, v_all_first_)";
        }
      };


      //
      //
      struct Complex: Traversal::Complex,
                      protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              parser_accessor_ (c),
              parser_base_set_ (c),
              parser_member_set_ (c),
              particle_ (c)
        {
          names_parser_accessor_ >> parser_accessor_;
          inherits_parser_base_set_ >> parser_base_set_;
          names_parser_member_set_ >> parser_member_set_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean he (has<Traversal::Element> (c));
          Boolean ha (has<Traversal::Attribute> (c));

          Boolean hae (has_particle<Traversal::Any> (c));
          Boolean haa (has<Traversal::AnyAttribute> (c));

          Boolean hra (false); // Has required attribute.
          if (ha)
          {
            RequiredAttributeTest test (hra);
            Traversal::Names names_test (test);
            names (c, names_test);
          }

          if (!((!restriction_p (c) && (he || ha)) ||
                validation && (he || hae || hra)))
            return;

          String const& name (ename (c));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          if (!restriction_p (c) && (he || ha))
          {
            // <name>_parser ()
            //
            names (c, names_parser_accessor_);


            // parsers ()
            //

            os << inl
               << "void " << name << "::" << endl
               << "parsers (";

            {
              ParserParamDecl decl (*this, true);
              decl.traverse (c);
            }

            os << ")"
               << "{";

            inherits (c, inherits_parser_base_set_);
            names (c, names_parser_member_set_);

            os << "}";
          }

          // Default c-tor.
          //
          os << inl
             << name << "::" << endl
             << name << " ()" << endl
             << ": ";

          Boolean comma (false);

          if (!restriction_p (c) && (he || ha))
          {
            ParserMemberInit member_init (*this);
            Traversal::Names names_member_init (member_init);

            names (c, names_member_init);

            comma = member_init.comma ();
          }

          if (validation && (he || hae))
          {
            if (comma)
              os << "," << endl << "  ";

            os << "v_state_stack_ (sizeof (v_state_), &v_state_first_)";

            particle_.dispatch (c.contains_compositor ().compositor ());

            comma = true;
          }

          if (validation && (hra))
          {
            if (comma)
              os << "," << endl << "  ";

            os << "v_state_attr_stack_ (sizeof (v_state_attr_), " <<
              "&v_state_attr_first_)";
          }

          os << "{"
             << "}";
        }

      private:
        //
        //
        ParserAccessor parser_accessor_;
        Traversal::Names names_parser_accessor_;

        //
        //
        ParserBaseSet parser_base_set_;
        Traversal::Inherits inherits_parser_base_set_;

        //
        //
        ParserMemberSet parser_member_set_;
        Traversal::Names names_parser_member_set_;

        //
        //
        Particle particle_;
      };
    }

    Void
    generate_parser_inline (Context& ctx)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      List list (ctx);
      Complex complex (ctx);

      names >> complex;
      names >> list;

      schema.dispatch (ctx.schema_root);
    }
  }
}

